# coding=utf-8
# Copyright 2014 Foursquare Labs Inc. All Rights Reserved.

from __future__ import absolute_import

from itertools import chain


class SymbolTreeNode(object):
  """A simple prefix tree with some custom logic for working with JVM symbols.

  Also used with Python symbols.
  """

  def __init__(self, children=None, values=None):
    self.children = children or {}
    self.values = values or set()

  def _insert(self, symbol_list, value):
    if not symbol_list:
      self.values.add(value)
    else:
      head, rest = symbol_list[0], symbol_list[1:]
      if head not in self.children:
        self.children[head] = SymbolTreeNode()
      child_node = self.children[head]
      child_node._insert(rest, value)

  def insert(self, symbol, value):
    """Insert value into the tree at symbol.

    :param string symbol: This symbol will be split on '.' to create nodes in the prefix tree.
    :param string value: This value (generally a source path) will be inserted into the set of
      values provided at the node produced by `symbol`.
    """
    return self._insert(symbol.split('.'), value)

  def flattened_subtree(self):
    """Returns the set of all values provided by this node and all of its descendents."""
    return set(chain(self.values,
                     *[child.flattened_subtree() for child in self.children.values()]))

  def _get(self, symbol_list, allow_prefix_imports, exact):
    if not symbol_list:
      if not self.values and allow_prefix_imports:
        return self.flattened_subtree()
      else:
        return set(self.values)
    elif symbol_list == ['_']:
      return self.flattened_subtree()
    else:
      head, rest = symbol_list[0], symbol_list[1:]
      if head in self.children:
        return self.children[head]._get(rest, allow_prefix_imports, exact)
      elif exact:
        return set()
      else:
        return set(self.values)

  def get(self, symbol, allow_prefix_imports=False, exact=False):
    """Returns the values at the node inferred from `symbol`.

    :param string symbol: This symbol will be split on '.' to traverse nodes in the prefix tree.
    :param bool allow_prefix_imports: If the symbol refers to a node which does not have any
      values associated with it (it's internal, and therefore has children), instead of returning
      nothing, return all of the values provided by its children (flattened_subtree).
    :param bool exact: Only return the values at a node if it is an exact match for the symbol.
      If a prefix node has values but does not have children to match the rest of the symbol,
      it will return an empty set.  Note that `allow_prefix_imports` and `exact` are mutually
      exclusive, and passing them both will cause an exception to be raised.
    """
    if allow_prefix_imports and exact:
      raise ValueError('Cannot call SymbolTreeNode.get() with both allow_prefix_imports=True and'
                       ' exact=True.  Symbole was: {0}'.format(symbol))
    return self._get(
      symbol.split('.'),
      allow_prefix_imports=allow_prefix_imports,
      exact=exact,
    )



# coding=utf-8
# Copyright 2016 Foursquare Labs Inc. All Rights Reserved.

from __future__ import (
  absolute_import,
  division,
  generators,
  nested_scopes,
  print_function,
  unicode_literals,
  with_statement,
)

from pants.task.task import Task
from pants.util.memo import memoized_property

from pants.contrib.buildgen.core.subsystems.buildgen_subsystem import BuildgenSubsystem


class BuildgenBase(Task):
  """"A base task that provides the buildgen subsystem to its implementers."""

  @classmethod
  def global_subsystems(cls):
    return super(BuildgenBase, cls).global_subsystems() + (BuildgenSubsystem.Factory,)

  @memoized_property
  def buildgen_subsystem(self):
    # TODO(pl): When pants is a proper library dep, remove this ignore.
    # pylint: disable=no-member
    return BuildgenSubsystem.Factory.global_instance().create()



# coding=utf-8
# Copyright 2013 Foursquare Labs Inc. All Rights Reserved.

from __future__ import (
  absolute_import,
  division,
  generators,
  nested_scopes,
  print_function,
  unicode_literals,
  with_statement,
)

import os
import time

from pants.base.build_environment import get_buildroot
from pants.task.task import Task


class BuildgenTimestamp(Task):
  """Log when buildgen was last ran in a findable place"""

  def execute(self):
    with open(os.path.join(get_buildroot(), '.buildgen_timestamp'), 'w') as f:
      f.write(str(int(time.time())))



# coding=utf-8
# Copyright 2016 Foursquare Labs Inc. All Rights Reserved.

from __future__ import (
  absolute_import,
  division,
  generators,
  nested_scopes,
  print_function,
  unicode_literals,
  with_statement,
)

import re

from pants.build_graph.target import Target


class BuildgenTargetBag(Target):
  """Aggregate all targets under a source tree directory that match the configured target type.

  BuildgenTargetBags targets have their dependencies managed by buildgen. Each target minimally defines a
  source directory and a target_alias type (e.g. 'python_library'). Buildgen will add a dependency to the implementing
  buildgen_target_bag BUILD definition for every matching target under that source dir.
  """

  def __init__(self,
               target_type_alias=None,
               source_tree=None,
               additional_generated_targets=None,
               ignored_targets_regex=None,
               *args,
               **kwargs):
    """
    :param str target_type_alias binary: The target_alias to gather into aggregated bag.
        (e.g. python_library or scala_library).
    :param str source_tree: Path relative to the buildroot - buildgen will gather the aggregated target_aliases
        from this dir.
    :param str additional_generated_targets: Any additional targets of the same target_alias that should be aggreagated
         but may not be under the same source_tree.
    :param str ignored_targets_regex: A regular expression capable of being compiled by the python 're' module.
        Targets whose address matches the regex will not be aggregated into the bag.
    """
    self.target_type_alias = target_type_alias
    self.source_tree = source_tree
    self.additional_generated_targets = additional_generated_targets
    self.ignored_targets_regex = re.compile(ignored_targets_regex) if ignored_targets_regex else None
    super(BuildgenTargetBag, self).__init__(*args, **kwargs)



# coding=utf-8
# Copyright 2014 Foursquare Labs Inc. All Rights Reserved.

from __future__ import (
  absolute_import,
  division,
  generators,
  nested_scopes,
  print_function,
  unicode_literals,
  with_statement,
)

from pants.base.build_environment import get_buildroot
from pants.base.cmd_line_spec_parser import CmdLineSpecParser
from pants.base.specs import DescendantAddresses

from pants.contrib.buildgen.core.buildgen_task import BuildgenTask


class Buildgen(BuildgenTask):

  @classmethod
  def alternate_target_roots(cls, options, address_mapper, build_graph):
    # NOTE(pl): This hack allows us to avoid sprinkling dummy tasks into every goal that we
    # need to be sure has its target roots modified.
    targets_required_downstream = set()

    buildgen_dirs = options.test_dirs + options.source_dirs
    source_dirs = [DescendantAddresses(d) for d in buildgen_dirs]

    for address in address_mapper.scan_specs(source_dirs):
      build_graph.inject_address_closure(address)
      targets_required_downstream.add(build_graph.get_target(address))
    return targets_required_downstream

  def execute(self):
    # NOTE(pl): We now rely on the fact that we've scheduled Buildgen (the dummy task in the
    # buildgen goal) to run before the real buildgen tasks, e.g. buildgen-scala, buildgen-thrift,
    # etc. Since we are being run before the real tasks but after everything else upstream,
    # we can fix the target roots back up to be whatever the buildgen tasks are supposed to
    # operate on (instead of the entire build graph, which the upstream operated on).

    build_graph = self.context.build_graph
    bg_target_roots = set()

    # NOTE(mateo): Using all source roots adds a scan of 3rdparty - may have a minor perf penalty.
    # I would like to switch to the configured source roots if I can find a way to surface it everywhere.
    buildgen_dirs = self.buildgen_subsystem.source_dirs + self.buildgen_subsystem.test_dirs
    all_buildgen_specs = ['{}::'.format(d) for d in buildgen_dirs]

    spec_parser = CmdLineSpecParser(get_buildroot())
    target_specs = self.context.options._target_specs or all_buildgen_specs
    parsed_specs = [
      spec_parser.parse_spec(target_spec)
      for target_spec in target_specs
    ]
    for address in self.context.address_mapper.scan_specs(parsed_specs):
      build_graph.inject_address_closure(address)
      bg_target_roots.add(build_graph.get_target(address))
    # Disable building from SCM for now, and instead always build everything unless the user
    # specifies less.
    # bg_target_roots.update(
    #   target for target in
    #   build_graph.transitive_dependees_of_addresses(t.address for t in self.targets_from_scm())
    #   if target.is_original
    # )
    self.context._replace_targets(list(bg_target_roots))



# coding=utf-8
# Copyright 2013 Foursquare Labs Inc. All Rights Reserved.

from __future__ import (
  absolute_import,
  division,
  generators,
  nested_scopes,
  print_function,
  unicode_literals,
  with_statement,
)

from pants.base.exceptions import TaskError
from pants.base.specs import DescendantAddresses
from pants.build_graph.address import Address

from pants.contrib.buildgen.core.buildgen_target_bag import BuildgenTargetBag
from pants.contrib.buildgen.core.buildgen_task import BuildgenTask


class BuildgenAggregateTargets(BuildgenTask):
  """Creates target bags that contain an aggregate of all matching target_types found under a specific source tree."""

  def __init__(self, context, workdir):
    super(BuildgenAggregateTargets, self).__init__(context, workdir)
    self._target_alias_whitelist = {'buildgen_target_bag'}

  def generate_target_type(self, build_graph, target_alias, source_tree, target_bag, additional_generated_targets=None,
                           ignored_targets_regex=None):
    extra_targets = additional_generated_targets or []
    generated_deps = {Address.parse(t) for t in extra_targets}

    for address in self.context.address_mapper.scan_specs([DescendantAddresses(source_tree)]):
      if not ignored_targets_regex or not ignored_targets_regex.match(address.spec):
        build_graph.inject_address_closure(address)
        target = build_graph.get_target(address)
        if address.spec_path != source_tree and target.type_alias == target_alias:
          generated_deps.add(address)

    self.adjust_target_build_file(
      build_graph.get_target(Address.parse(target_bag)),
      generated_deps,
      whitelist=self._target_alias_whitelist,
    )

  def execute(self):
    build_graph = self.context.build_graph
    target_bags = []
    # For each BuildgenTargetBag, gather the matching targets and aggregate them as dependencies of the bag target.
    for spec in self.buildgen_subsystem.buildgen_target_bags:
      address = Address.parse(spec)
      build_graph.inject_address_closure(address)
      bag_target = build_graph.get_target(address)
      if not isinstance(bag_target, BuildgenTargetBag):
        raise TaskError("The generated_target_bags option must point to a list of BuildgenTargetBag specs.")
      target_bags.append(bag_target)

    for target in target_bags:
      self.generate_target_type(
        build_graph,
        target_alias=target.target_type_alias,
        source_tree=target.source_tree,
        additional_generated_targets=target.additional_generated_targets,
        target_bag=target.address.spec,
        ignored_targets_regex=target.ignored_targets_regex,
      )



# coding=utf-8
# Copyright 2013 Foursquare Labs Inc. All Rights Reserved.

from __future__ import (
  absolute_import,
  division,
  generators,
  nested_scopes,
  print_function,
  unicode_literals,
  with_statement,
)

from copy import deepcopy

from pants.util.memo import memoized_property

from pants.contrib.buildgen.core.build_file_manipulator import BuildFileManipulator
from pants.contrib.buildgen.core.buildgen_base import BuildgenBase
from pants.contrib.buildgen.core.third_party_map_util import merge_map


class BuildgenTask(BuildgenBase):

  @classmethod
  def prepare(cls, options, round_manager):
    round_manager.require('concrete_target_to_derivatives')
    round_manager.require('source_to_addresses_mapper')

  @memoized_property
  def dryrun(self):
    return self.buildgen_subsystem.dry_run

  @memoized_property
  def target_alias_whitelist(self):
    return self.buildgen_subsystem.target_alias_whitelist

  @memoized_property
  def target_alias_blacklist(self):
    """Subclasses may implement to list target aliases that should not be managed by that task.

    For example, a buildgem task may understand JvmLibrary but not its subclass ScalaLibrary.
    """
    return []

  @memoized_property
  def managed_dependency_aliases(self):
    return self.buildgen_subsystem.managed_dependency_aliases

  @memoized_property
  def third_party_target_aliases(self):
    """List of target_aliases that are acceptable 3rdparty libraries for a language."""
    return []

  @memoized_property
  def merged_map(self):
    """Returns the recursively updated mapping of imports to third party BUILD file entries.

    Entries passed to the option system take priority.
    """
    # TODO(mateo): Make the third_party map a Task property and move this to super class?
    merged_map = deepcopy(self.third_party_map)
    merge_map(merged_map, self.get_options().additional_third_party_map)
    return merged_map

  def adjust_target_build_file(self, target, computed_dep_addresses, whitelist=None):
    """Makes a BuildFileManipulator and adjusts the BUILD file to reflect the computed addresses"""
    alias_whitelist = whitelist or self.buildgen_subsystem.target_alias_whitelist
    manipulator = BuildFileManipulator.load(target.address.build_file,
                                            target.address.target_name,
                                            alias_whitelist)

    existing_dep_addresses = manipulator.get_dependency_addresses()
    for address in existing_dep_addresses:
      if not self.context.build_graph.get_target(address):
        self.context.build_graph.inject_address(address)

    existing_deps = [self.context.build_graph.get_target(address)
                     for address in existing_dep_addresses]
    ignored_deps = [dep for dep in existing_deps
                    if dep.type_alias not in self.managed_dependency_aliases]

    manipulator.clear_unforced_dependencies()
    for ignored_dep in ignored_deps:
      manipulator.add_dependency(ignored_dep.address)
    for address in computed_dep_addresses:
      manipulator.add_dependency(address)

    final_dep_addresses = manipulator.get_dependency_addresses()
    manipulator.write(dry_run=self.dryrun, use_colors=self.get_options().colors)

  def execute(self):
    def task_targets():
      for target in self.context.target_roots:
        if isinstance(target, self.types_operated_on) and target.type_alias not in self.target_alias_blacklist:
          yield target
    targets = sorted(list(task_targets()))
    if self.get_options().level == 'debug':
      print('\n{0} will operate on the following targets:'.format(type(self).__name__))
      for target in targets:
        print('* {0}'.format(target.address.reference()))
    for target in targets:
      self.buildgen_target(target)



# coding=utf-8
# Copyright 2014 Foursquare Labs Inc. All Rights Reserved.

from __future__ import absolute_import

from pants.build_graph.build_file_aliases import BuildFileAliases
from pants.goal.task_registrar import TaskRegistrar as task

from pants.contrib.buildgen.core.buildgen import Buildgen
from pants.contrib.buildgen.core.buildgen_aggregate_targets import BuildgenAggregateTargets
from pants.contrib.buildgen.core.buildgen_target_bag import BuildgenTargetBag
from pants.contrib.buildgen.core.buildgen_timestamp import BuildgenTimestamp
from pants.contrib.buildgen.core.map_derived_targets import MapDerivedTargets
from pants.contrib.buildgen.core.map_sources_to_addresses import MapSourcesToAddresses


def build_file_aliases():
  return BuildFileAliases(
    targets={
      'buildgen_target_bag': BuildgenTargetBag,
    },
  )

def register_goals():

  task(
    name='map-derived-targets',
    action=MapDerivedTargets,
  ).install()

  task(
    name='map-sources-to-addresses-mapper',
    action=MapSourcesToAddresses,
  ).install()

  task(
    name='buildgen',
    action=Buildgen,
  ).install()

  task(
    name='aggregate-targets',
    action=BuildgenAggregateTargets,
  ).install('buildgen')

  task(
    name='timestamp',
    action=BuildgenTimestamp,
  ).install('buildgen')



# coding=utf-8
# Copyright 2014 Foursquare Labs Inc. All Rights Reserved.

from __future__ import (
  absolute_import,
  division,
  generators,
  nested_scopes,
  print_function,
  unicode_literals,
  with_statement,
)

from pants.build_graph.source_mapper import LazySourceMapper
from pants.task.task import Task


class MapSourcesToAddresses(Task):
  """Map sources to addresses.

  Tasks that buildgen for a target type can extend this task's product with their own source -> address mapping.
  """

  @classmethod
  def product_types(cls):
    return [
      'source_to_addresses_mapper',
    ]

  @classmethod
  def prepare(cls, options, round_manager):
    # Not JVM specific but needed to ensure codegen is run before buildgen.
    round_manager.require_data('java')
    round_manager.require_data('scala')

  def execute(self):
    products = self.context.products
    source_to_addresses_mapper = LazySourceMapper(
      self.context.address_mapper,
      self.context.build_graph,
      stop_after_match=True,
    )
    for target in self.context.build_graph.targets():
      for source in target.sources_relative_to_buildroot():
        source_to_addresses_mapper._source_to_address[source].add(target.address)
    products.safe_create_data('source_to_addresses_mapper',
                              lambda: source_to_addresses_mapper)



# coding=utf-8
# Copyright 2016 Foursquare Labs Inc. All Rights Reserved.

from __future__ import absolute_import


def merge_map(a, b):
  """Recursively merge two nested dictionaries, with any given subtree in b recursively taking priority over a."""
  # Callers are encouraged to pass a deepcopy of dict a.
  # TODO(mateo): Remove mutation and return a new dict instead.

  # Recursion terminates on any type that isn't a dictionary, which means that values in b can clobber
  # entire subtrees of a. I consider that to be the correct thing to do.

  if not isinstance(a, dict) or not isinstance(b, dict):
    return b

  # If a b subtree defines a 'DEFAULT' then assume that it maps the world of b's acceptable mappings for that subtree.
  if 'DEFAULT' in b:
    a.clear()
  for key in b.keys():
    a[key] = merge_map(a[key], b[key]) if key in a else b[key]
  return a


def check_manually_defined(symbol, subtree=None):
  """Checks to see if a symbol was hand mapped by third_party_map."""

  if subtree == 'SKIP':
    return 'SKIP'
  if isinstance(subtree, basestring) or subtree is None:
    return subtree

  parts = symbol.split('.')
  if not parts:
    raise ValueError('Came to the end of the symbol while still traversing tree.  Subtree '
                     'keys were {keys}'.format(keys=subtree.keys()))
  if parts[0] not in subtree:
    if 'DEFAULT' in subtree:
      return subtree['DEFAULT']
    else:
      return None
  else:
    return check_manually_defined('.'.join(parts[1:]), subtree[parts[0]])



# coding=utf-8
# Copyright 2016 Foursquare Labs Inc. All Rights Reserved.

from __future__ import (
  absolute_import,
  division,
  generators,
  nested_scopes,
  print_function,
  unicode_literals,
  with_statement,
)

from pants.subsystem.subsystem import Subsystem


class BuildgenSubsystem(object):
  """Subsystem for Buildgen global configuration."""

  class Factory(Subsystem):
    options_scope = 'buildgen'

    @classmethod
    def register_options(cls, register):
      # TODO(mateo): Is there a way to surface the configured Source Roots here? Not sure if we have that sort
      # of guarantee about option loading - although you might think that source_roots would be known by this point.
      register(
        "--source-dirs",
        default=['src', '3rdparty'],
        advanced=True,
        type=list,
        help="Source directory that holds targets to buildgen - defaults to the configured source roots.",
      )
      register(
        '--test-dirs',
        default=['test'],
        advanced=True,
        type=list,
        help="Source directory that holds targets to buildgen - defaults to the configured source roots.",
      )
      register(
        '--target-whitelist',
        default=['target'],
        advanced=True,
        type=list,
        help='Full list of target callables that are allowed to be injected into the build file context.',
      )
      register(
        '--managed-dependency-aliases',
        advanced=True,
        type=list,
        help='The superset of target aliases that buildgen can inject as dependencies '
             '(e.g. [scala_library, python_library, ...].',
      )
      register(
        '--buildgen-target-bags',
        default=[],
        advanced=True,
        type=list,
        help="Each listed BuildgenTargetBag will have all configured targets aggregated and injected as dependencies",
      )
      register(
        '--dry-run',
        default=False,
        type=bool,
        help='When True, buildgen will rewrite BUILD files in-place, otherwise it will print a diff to stdout.'
      )

    def create(self):
      options = self.get_options()
      return BuildgenSubsystem(
        options.source_dirs,
        options.test_dirs,
        options.target_whitelist,
        options.managed_dependency_aliases,
        options.buildgen_target_bags,
        options.dry_run
      )

  def __init__(self,
               source_dirs,
               test_dirs,
               target_whitelist,
               managed_dependency_aliases,
               buildgen_target_bags,
               dry_run):
    self.source_dirs = source_dirs
    self.test_dirs = test_dirs
    self.target_alias_whitelist = target_whitelist
    self.managed_dependency_aliases = managed_dependency_aliases
    self.buildgen_target_bags = buildgen_target_bags
    self.dry_run = dry_run




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: ContribPlugin(BuildFileAddress(BuildFile(contrib/buildgen/src/python/pants/contrib/buildgen/core/BUILD, FileSystemProjectTree(/Users/mateo/dev/pants)), plugin))

from setuptools import setup

setup(**
{   'classifiers': [   'Intended Audience :: Developers',
                       'License :: OSI Approved :: Apache Software License',
                       'Operating System :: MacOS :: MacOS X',
                       'Operating System :: POSIX :: Linux',
                       'Programming Language :: Python',
                       'Topic :: Software Development :: Build Tools',
                       'Topic :: Software Development :: Code Generators'],
    'description': "Foursquare's Buildgen Pants plugin core module",
    'entry_points': {   'pantsbuild.plugin': [   'build_file_aliases = pants.contrib.buildgen.core.register:build_file_aliases',
                                                 'register_goals = pants.contrib.buildgen.core.register:register_goals']},
    'install_requires': ['ansicolors==1.0.2', 'pantsbuild.pants>=1.0.0,<1.5.0'],
    'license': 'Apache License, Version 2.0',
    'long_description': None,
    'name': 'fsqio.pants.contrib.buildgen.core',
    'namespace_packages': [   'pants',
                              'pants.contrib',
                              'pants.contrib.buildgen',
                              'pants.contrib.buildgen.core',
                              'pants.contrib.buildgen.core.subsystems'],
    'package_data': {   },
    'package_dir': {   '': 'src'},
    'packages': [   'pants',
                    'pants.contrib',
                    'pants.contrib.buildgen',
                    'pants.contrib.buildgen.core',
                    'pants.contrib.buildgen.core.subsystems'],
    'url': 'https://github.com/foursquare/fsqio',
    'version': '1.1.0',
    'zip_safe': True}
)



# coding=utf-8
# Copyright 2014 Foursquare Labs Inc. All Rights Reserved.

from __future__ import (
  absolute_import,
  division,
  generators,
  nested_scopes,
  print_function,
  unicode_literals,
  with_statement,
)

from collections import defaultdict

from pants.task.task import Task


class MapDerivedTargets(Task):
  """Provides a product mapping concrete targets to all of their synthetic derivatives."""

  @classmethod
  def product_types(cls):
    return [
      'concrete_target_to_derivatives',
    ]

  @classmethod
  def prepare(cls, options, round_manager):
    # It is not necessarily JVM specific but it must be scheduled after codegen.
    round_manager.require_data('java')
    round_manager.require_data('python')
    round_manager.require_data('scala')

  def execute(self):
    concrete_target_to_derivatives = defaultdict(set)
    for target in self.context.build_graph.targets():
      if target.is_synthetic and target.derived_from is not None:
        concrete_target_to_derivatives[target.concrete_derived_from].add(target)
    products = self.context.products
    products.safe_create_data('concrete_target_to_derivatives',
                              lambda: concrete_target_to_derivatives)



# coding=utf-8
# Copyright 2014 Foursquare Labs Inc. All Rights Reserved.

from __future__ import (
  absolute_import,
  division,
  generators,
  nested_scopes,
  print_function,
  unicode_literals,
  with_statement,
)

from itertools import chain
import json
import os
from os import path as P

from pants.base.build_environment import get_buildroot
from pants.task.task import Task


class SourceAnalysisTask(Task):
  """A parent Task for buildgen upstream analysis tasks which provide source -> analysis products.

  SourceAnalysisTask factors out a large amount of boilerplate and handles caching logic.
  """

  @classmethod
  def product_types(cls):
    return [
      cls.analysis_product_name(),
    ]

  @classmethod
  def analysis_product_name(cls):
    """A string that names the product type this task produces."""
    raise NotImplementedError()

  @property
  def claimed_target_types(self):
    """A tuple of target types that the implementing class will do source analysis on."""
    raise NotImplementedError()

  def is_analyzable(self, source_relpath):
    """Whether or not a source is a candidate for analysis.

    Generally this will just be a test of the source's extension.
    """
    raise NotImplementedError()

  def analyze_sources(self, source_relpaths):
    """Called to analyze uncached sources.

    Returns a map from source relpaths to opaque analysis objects.
    It is an error f the returned map does not have a key for every passed in source relpath.
    """
    raise NotImplementedError()

  def __init__(self, *args, **kwargs):
    super(SourceAnalysisTask, self).__init__(*args, **kwargs)

  def execute(self):
    products = self.context.products
    targets = [t for t in self.context.target_roots if isinstance(t, self.claimed_target_types)]

    def vt_analysis_file(vt):
      return P.join(self._workdir, vt.target.id, vt.cache_key.hash, 'analysis.json')

    with self.invalidated(targets, invalidate_dependents=False) as invalidation_check:
      analyzable_sources_by_target = {
        vt.target: set(
          source for source in vt.target.sources_relative_to_buildroot()
          if self.is_analyzable(source)
        )
        for vt in invalidation_check.all_vts
      }
      invalid_targets = [vt.target for vt in invalidation_check.invalid_vts]
      invalid_analyzable_sources = set(
        chain.from_iterable(
          [source for source in t.sources_relative_to_buildroot() if self.is_analyzable(source)]
          for t in invalid_targets
        )
      )
      calculated_analysis = self.analyze_sources(list(invalid_analyzable_sources))
      uncalculated_analysis = invalid_analyzable_sources - set(calculated_analysis.keys())
      if uncalculated_analysis:
        raise Exception(
          '{0} failed to calculate analysis for the following sources:\n *{1}\n'
          .format(
            type(self).__name__,
            '\n *'.join(sorted(uncalculated_analysis))
          )
        )

      vts_artifactfiles_pairs = []
      for vt in invalidation_check.invalid_vts:
        target_analysis = {}
        for source in analyzable_sources_by_target[vt.target]:
          target_analysis[source] = calculated_analysis[source]
        target_analysis_bytes = json.dumps(target_analysis)
        target_analysis_path = vt_analysis_file(vt)
        target_analysis_abs_dir = P.dirname(P.join(get_buildroot(), target_analysis_path))
        if not P.isdir(target_analysis_abs_dir):
          os.makedirs(target_analysis_abs_dir)
        with open(P.join(get_buildroot(), target_analysis_path), 'wb') as f:
          f.write(target_analysis_bytes)
        vts_artifactfiles_pairs.append((vt, [target_analysis_path]))

      analysis_product = {}
      analysis_product.update(calculated_analysis)

      for vt in invalidation_check.all_vts:
        if vt.valid:
          target_analysis_path = vt_analysis_file(vt)
          with open(P.join(get_buildroot(), target_analysis_path), 'rb') as f:
            analysis_bytes = f.read()
          target_analysis = json.loads(analysis_bytes)
          analysis_product.update(target_analysis)

      if self.artifact_cache_writes_enabled():
        self.update_artifact_cache(vts_artifactfiles_pairs)
      products.safe_create_data(self.analysis_product_name(), lambda: analysis_product)
